#include <assert.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <memory>
#include <stack>
#include <string>
#include <vector>

#include <pcre.h>

using namespace std;

int g_verbose = 0;

void error(const char* fmt, ...) {
    va_list va;
    va_start(va, fmt);
    vfprintf(stderr, fmt, va);
    va_end(va);
    fputc('\n', stderr);
    exit(1);
}

void log(const char* fmt, ...) {
    if (g_verbose) {
        fputs("WAKE: ", stderr);
        va_list va;
        va_start(va, fmt);
        vfprintf(stderr, fmt, va);
        va_end(va);
        fputc('\n', stderr);
    }
}

class Str {
public:
    Str() : ptr_(NULL), len_(0), reserved_(0) {
    }

    Str(const char* p) {
        init(p, strlen(p));
    }

    Str(const char* p, int len) {
        init(p, len);
    }

    Str(const Str& s) {
        init(s.ptr_, s.len_);
    }

    ~Str() {
        free(ptr_);
    }

    char operator[](int i) const {
        return ptr_[i];
    }

    bool operator==(const Str& s) const {
        if (len_ != s.len_) {
            return false;
        }
        return !memcmp(ptr_, s.ptr_, len_);
    }

    Str operator+(const Str& s) const {
        Str r(*this);
        r += s;
        return r;
    }

    Str& operator+=(const Str& s) {
        insert(end(), s.ptr_, s.len_);
        return *this;
    }

    Str& operator+=(const char* p) {
        insert(end(), p, strlen(p));
        return *this;
    }

    const char* begin() const {
        return ptr_;
    }

    const char* end() const {
        return ptr_ + len_;
    }

    void insert(const char* pos, const char* p, int len) {
        assert(pos == end());
        reserve(len_ + len + 1);
        memcpy(ptr_ + len_, p, len);
        len_ += len;
        ptr_[len_] = '\0';
    }

    bool empty() const {
        return !len_;
    }

    Str substr(int o, int s) const {
        return Str(ptr_ + o, s);
    }

    void clear() {
        if (ptr_) {
            len_ = 0;
            ptr_[0] = '\0';
        }
    }

    unsigned size() const {
        return len_;
    }

    const char* c_str() const {
        return ptr_;
    }

    const char* data() const {
        return ptr_;
    }

    void reserve(int s) {
        if (reserved_ < s) {
            ptr_ = (char*)realloc(ptr_, s);
            reserved_ = s;
        }
    }

    void push_back(char c) {
        if (len_ + 1 >= reserved_) {
            if (len_) {
                reserve(len_ * 3 / 2);
            } else {
                reserve(10);
            }
        }
        ptr_[len_++] = c;
        ptr_[len_] = '\0';
    }

private:
    void init(const char* p, int len) {
        len_ = len;
        reserved_ = len_ + 1;
        ptr_ = (char*)malloc(reserved_);
        memcpy(ptr_, p, len_);
        ptr_[len_] = '\0';
    }

    char* ptr_;
    int len_;
    int reserved_;
};

#define string Str

class Rule {
public:
    struct Action {
        bool is_str;
        string act;
        int lineno;
    };

    ~Rule() {
        pcre_free(regexp_);
    }

    static void parse(const char* wakefile,
                      int from,
                      vector<Rule*>* rules) {
        string line;
        int lineno = 1;
        FILE* fp = fopen(wakefile, "rb");
        if (!fp) {
            if (from == -1) {
                error("%s: no such file", wakefile);
            } else {
                error("%s: no such file in line %d", wakefile, from);
            }
        }

        int num_rules = 0;
        int num_acts = 0;
        while (true) {
            int c = fgetc(fp);
            if (c == '\n' || c == '\r' || c == EOF) {
                if (!line.empty()) {
                    if (line[0] == '#') {
                        if (line.substr(0, strlen("#include")) == "#include") {
                            parseInclude(line, lineno, rules);
                        }
                    } else {
                        Rule* rule = new Rule(line, lineno);
                        num_rules++;
                        num_acts += rule->acts().size();
                        rules->push_back(rule);
                    }
                }
                if (c == EOF) {
                    break;
                }

                if (c == '\n') {
                    lineno++;
                }
                line.clear();
            } else {
                line.push_back(c);
            }
        }
        fclose(fp);

        log("%s - %d rules and %d actions in %d lines",
            wakefile, num_rules, num_acts, lineno - 1);
    }

    const string& target() const { return target_; }
    const pcre* regexp() const { return regexp_; }
    const vector<Action>& acts() const { return acts_; }
    int lineno() const { return lineno_; }

private:
    Rule(const string& line, int lineno)
        : lineno_(lineno) {
        size_t i = parseTarget(line);
        parseActions(line, i);
    }

    static void parseInclude(const string& line, int lineno,
                             vector<Rule*>* rules) {
        size_t i = strlen("#include");
        while (true) {
            if (i + 2 >= line.size()) {
                error("invalid #include in line %d: %s", lineno, line.c_str());
            }
            if (!isspace(line[i])) {
                break;
            }
            i++;
        }
        if (line[i] != '"' || line[line.size() - 1] != '"') {
            error("invalid #include in line %d: %s", lineno, line.c_str());
        }
        parse(line.substr(i + 1, line.size() - i - 2).c_str(), lineno, rules);
    }

    size_t parseTarget(const string& line) {
        size_t i = 0;
        while (true) {
            if (i == line.size()) {
                error("No colons in line %d", lineno_);
            }

            if (line[i] == ':') {
                i++;
                break;
            }

            if (line[i] == '\\' && i+1 < line.size() && line[i+1] == ':') {
                ++i;
            }
            target_.push_back(line[i]);

            ++i;
        }

        const char* err;
        int offset;
        regexp_ = pcre_compile((string("\\A(?:") + target_ + ")\\Z").c_str(),
                               PCRE_MULTILINE | PCRE_DOTALL,
                               &err, &offset, NULL /* tableptr */);
        if (!regexp_) {
            error("Regexp error: %s \"%s\" in line %d at %d",
                  err, target_.c_str(), lineno_, offset - 5);
        }

        return i;
    }

    void parseActions(const string& line, size_t i) {
        while (i < line.size()) {
            while (isspace(line[i])) {
                i++;
                if (i == line.size()) {
                    return;
                }
            }
            if (line[i] == '#') {
                return;
            }

            Action act;
            act.lineno = lineno_;
            if (line[i] == '"') {
                act.is_str = true;
                for (i++;; i++) {
                    if (i == line.size()) {
                        error("Unmatched \" in line %d action %d",
                              lineno_, acts_.size() + 1);
                    }
                    if (line[i] == '"') {
                        break;
                    } else if (line[i] == '\\') {
                        act.act.push_back(line[i]);
                        i++;
                    }
                    act.act.push_back(line[i]);
                }
                i++;
            } else {
                act.is_str = false;
                while (i < line.size() &&
                       line[i] != '"' && !isspace(line[i])) {
                    act.act.push_back(line[i]);
                    i++;
                    if (line[i-1] == '\\' && i < line.size()) {
                        act.act.push_back(line[i]);
                        i++;
                    }
                }
            }
            acts_.push_back(act);
        }
    }

    pcre* regexp_;
    string target_;
    vector<Action> acts_;
    int lineno_;
};

class Runner {
public:
    struct Matched {
    public:
        Matched() : subject(NULL), offsets(NULL), num_substr(0), refcnt(0) {}
        ~Matched() {
            delete subject;
            free(offsets);
        }

        void deref() {
            if (this) {
                refcnt--;
                if (refcnt == 0) {
                    delete this;
                }
            }
        }

        string* subject;
        int* offsets;
        int num_substr;
        int refcnt;
    };

    void run(const vector<Rule*>& rules,
             const vector<Rule::Action>& acts,
             Matched* matched,
             string* out) {
        typedef pair<const Rule::Action*, Matched*> Context;
        stack<Context, vector<Context> > st;
        for (int i = acts.size() - 1; i >= 0; i--) {
            st.push(Context(&acts[i], matched));
        }

        while (!st.empty()) {
            Context ctx = st.top();
            st.pop();

            const Rule::Action* act = ctx.first;
            Matched* matched = ctx.second;

            if (act->is_str) {
                string* p = out;
                string buf;
                if (!out) {
                    p = &buf;
                }
                expandString<true>(rules, act->act, matched, act->lineno, p);
                if (!out) {
                    fwrite(p->data(), 1, p->size(), stdout);
                    fflush(stdout);
                }
            } else {
                const vector<Rule::Action>* nacts = 0;
                Matched* nmatched = 0;
                lookupAction(rules, act->act, matched, act->lineno,
                             &nacts, &nmatched);
                for (int i = nacts->size() - 1; i >= 0; i--) {
                    st.push(Context(&(*nacts)[i], nmatched));
                }
            }

            matched->deref();
        }
    }

private:
    template <bool is_str>
    void expandString(const vector<Rule*>& rules,
                      const string str,
                      Matched* matched,
                      int lineno,
                      string* out) {
        out->reserve(out->size() + str.size());
        for (size_t i = 0; i < str.size(); i++) {
            if (is_str) {
            } else {
                const char* p = str.data() + i;
                const char* np = (const char*)memchr(p, '$', str.size() - i);
                if (np) {
                    out->insert(out->end(), p, np - p);
                    i += np - p;
                } else {
                    out->insert(out->end(), p, str.size() - i);
                    break;
                }
            }

            if (!is_str || str[i] == '$') {
                if (i + 1 < str.size()) {
                    char c = str[i+1];
                    int n;
                    switch (c) {
                    case '+':
                        for (n = matched->num_substr - 1; n >= 1; n--) {
                            if (matched->offsets[n * 2] >= 0) {
                                goto concat_matched;
                            }
                        }
                        i++;
                        goto done;
                    case '&':
                        c = '0';
                        // fall through
                    case '1':
                    case '2':
                    case '3':
                    case '4':
                    case '5':
                    case '6':
                    case '7':
                    case '8':
                    case '9': {
                        n = c - '0';
                        if (matched && n < matched->num_substr) {
                        concat_matched:
                            int* offset = matched->offsets + n * 2;
                            out->insert(out->end(),
                                        matched->subject->data() + offset[0],
                                        offset[1] - offset[0]);
                        }
                        i++;
                        goto done;
                    }
                    case '<': {
                        if (!stdin_.get()) {
                            stdin_.reset(new string);
                            char* buf = new char[1024];
                            size_t n;
                            while ((n = fread(buf, 1, 1024, stdin)) > 0) {
                                *stdin_ += string(buf, n);
                                if (n != 1024) {
                                    break;
                                }
                            }
                            delete[] buf;
                        }
                        *out += *stdin_;
                        i++;
                        goto done;
                    }
                    case '(': {
                        string act;
                        act.reserve(str.size() - i);
                        i += 2;
                        int num_open = 1;
                        while (true) {
                            if (i == str.size()) {
                                error("%s: unmatched paren in %d",
                                      str.c_str(), lineno);
                            }
                            if (str[i] == '(') {
                                num_open++;
                            } else if (str[i] == ')') {
                                num_open--;
                            }

                            if (!num_open) {
                                break;
                            }
                            act.push_back(str[i]);
                            i++;
                        }

                        log("%s @%d in %s", act.c_str(), lineno, str.c_str());
                        const vector<Rule::Action>* nacts = 0;
                        Matched* nmatched = 0;
                        lookupAction(rules, act, matched, lineno,
                                     &nacts, &nmatched);
                        run(rules, *nacts, nmatched, out);
                        goto done;
                    }
                    }
                }
            } else if (str[i] == '\\') {
                if (i + 1 < str.size()) {
                    i++;
                    if (str[i] == 'n') {
                        out->push_back('\n');
                        continue;
                    } else if (str[i] == 'r') {
                        out->push_back('\r');
                        continue;
                    }
                }
            }
            out->push_back(str[i]);
        done:
            ;
        }
    }

    void lookupAction(const vector<Rule*>& rules,
                      const string& orig_act,
                      Matched* matched,
                      int lineno,
                      const vector<Rule::Action>** out_acts,
                      Matched** out_matched) {
        string* act = new string;
        expandString<true>(rules, orig_act, matched, lineno, act);
        for (size_t i = 0; i < rules.size(); i++) {
            const Rule* rule = rules[i];
            int ovector[30];
            int count = pcre_exec(rule->regexp(),
                                  NULL  /* extra */,
                                  act->data(),
                                  act->size(),
                                  0  /* startoffset */,
                                  0  /* options */,
                                  ovector,
                                  30);
            if (count < 0) {
                continue;
            }

            log("%s @%d => %s @%d",
                act->c_str(), lineno, rule->target().c_str(), rule->lineno());

            *out_acts = &rule->acts();
            if (rule->acts().size()) {
                *out_matched = new Matched();
                (*out_matched)->subject = act;
                (*out_matched)->num_substr = count;
                assert(count);
                int bufsize = sizeof(int) * count * 2;
                (*out_matched)->offsets = (int*)malloc(bufsize);
                memcpy((*out_matched)->offsets, ovector, bufsize);
                (*out_matched)->refcnt += rule->acts().size();
            } else {
                delete act;
                *out_matched = NULL;
            }
            return;
        }
        error("%s: no rule", act->c_str());
    }

    auto_ptr<string> stdin_;
};

int main(int argc, char* argv[]) {
    const char* wakefile = NULL;
    for (int i = 1; i < argc; i++) {
        if (!strcmp(argv[i], "-v")) {
            g_verbose = 1;
        } else if (!wakefile) {
            wakefile = argv[i];
        }
    }

    if (!wakefile) {
        error("Usage: %s [-v] <wake-file>", argv[0]);
    }

    vector<Rule*> rules;
    Rule::parse(wakefile, -1, &rules);
    if (rules.empty()) {
        log("No rules");
        exit(0);
    }

    Runner runner;
    runner.run(rules, rules[0]->acts(), NULL, NULL);

    for (size_t i = 0; i < rules.size(); i++) {
        delete rules[i];
    }

    return 0;
}
