#!/usr/bin/env ruby

def error(msg)
  STDERR.puts msg
  exit 1
end

def log(msg)
  if $verbose
    STDERR.puts "WAKE: " + msg
  end
end

def parse_regexp(line, lineno)
  i = 0
  r = ''
  while line[i] != ?:
    if !line[i]
      error "No colons in line #{lineno}"
    end

    if line[i] == ?\\ && line[i+1] == ?:
      i += 1
    end
    r << line[i]

    i += 1
  end
  [r, line[i+1..-1]]
end

def parse_actions(line, lineno)
  actions = []

  while true
    line.sub!(/^\s*(?:#.*)?/, '')
    break if line.empty?

    if line[0] == ?"
      sa = [s = '']
      i = 1
      while line[i] != ?"
        if !line[i]
          error "Unmatched \" in line #{lineno} action #{actions.size+1}"
        end
        if line[i] == ?\\
          i += 1
          if line[i] == ?n
            s << ?\n
            i += 1
            next
          elsif line[i] == ?r
            s << ?\r
            i += 1
            next
          elsif line[i] == ?$
            sa << (s = '')
            i += 1
            next
          end
        end
        s << line[i]

        i += 1
      end

      line = line[i+1..-1]
      actions << [:str, sa, lineno]
    else
      line.sub!(/([^\s"]*(\\\s|\\"))*[^\s"]*/, '')
      actions << [:act, $&, lineno]
    end
  end

  actions
end

def parse(file, from = '')
  rules = []
  begin
    lines = File.readlines(file)
  rescue
    error("#{file}: no such file#{from}")
  end
  num_rules = num_actions = 0
  lines.each_with_index do |line, lineno|
    lineno += 1

    if line =~ /^#include\b/
      if $' !~ /^\s*"(.*)"$/
        error("invalid #include in line #{lineno}: #{line}")
      end
      rules += parse($1, " in line #{lineno}")
      next
    end

    line.chomp!
    next if line.empty? || line[0] == ?#

    regexp, line = parse_regexp(line, lineno)
    next if !regexp
    actions = parse_actions(line, lineno)

    num_rules += 1
    num_actions += actions.size

    reg = begin
            Regexp.new('\\A(?:' + regexp + ')\\Z', Regexp::MULTILINE)
          rescue
            error("Regexp error: #{$!.to_s} \"#{regexp}\" in line #{lineno}")
          end
    rules << [reg, actions, regexp, lineno]
  end

  log("#{file} - #{num_rules} rules and #{num_actions} actions " +
      "in #{lines.size} lines")

  rules
end

def ev(rules, str, matched, lineno, is_str)
  o = ''
  i = 0
  while str[i]
    if str[i] == ?$ && str[i+1]
      c = str[i+1]
      if mi = '&123456789'.index(c)
        o << matched[mi].to_s
        i += 2
        next
      elsif c == ?+
        mi = matched.size - 1
        while mi >= 1
          if matched[mi]
            o << matched[mi].to_s
            break
          end
          mi -= 1
        end
        i += 2
        next
      elsif c == ?<
        o << ($read_stdin ||= STDIN.read)
        i += 2
        next
      elsif c == ?(
        act = ''
        i += 2
        num_open = 1
        while num_open != 0
          c = str[i]
          if !c
            error "#{str}: unmatched paren in line #{lineno}"
          elsif c == ?(
            num_open += 1
          elsif c == ?)
            num_open -= 1
          end
          act << c
          i += 1
        end

        act.chop!
        log("#{act} @#{lineno} in #{str}")
        args = lookup_act(rules, act, matched, lineno)
        o << run(rules, *args)
        next
      end
    elsif !is_str && str[i] == ?\\ && str[i+1]
      i += 1
      if str[i] == ?n
        o << ?\n
        i += 1
        next
      elsif str[i] == ?r
        o << ?\r
        i += 1
        next
      end
    end
    o << str[i]
    i += 1
  end
  o
end

def lookup_act(rules, act, matched, lineno)
  act = ev(rules, act, matched, lineno, false)
  rules.any? do |reg, actions, orig_reg, reg_line|
    if reg =~ act
      log("#{act} @#{lineno} => #{orig_reg} @#{reg_line}")
      return [actions, $~.dup]
    end
  end
  error("#{act}: no rule")
end

def run(rules, acts, orig_matched, toplevel = false)
  out = ''
  i = 0
  while acts[i]
    op, act, lineno, matched = acts[i]
    matched = orig_matched if !matched
    i += 1

    if op == :str
      o = act.map{|a|ev(rules, a, matched, lineno, true)} * '$'
      if toplevel
        print o
      else
        out += o
      end
      next
    end

    # Tail call.
    nacts, nmatched = lookup_act(rules, act, matched, lineno)
    acts = nacts.map{|a|a + [nmatched]} + acts[i..-1]
    i = 0
  end
  out
end

$verbose = ARGV.delete('-v')

if !ARGV[0]
  error "Usage: #$0 [-v] <wake-file>"
end

rules = parse(ARGV[0])
#p rules

if rules.empty?
  log("No rules")
  exit
end

run(rules, rules[0][1], [], true)
